@@ -1,4 +1,10 @@
-{-# LANGUAGE OverloadedStrings #-}
+{-# OPTIONS_GHC -fno-warn-missing-methods #-}
+{-# LANGUAGE DeriveFunctor         #-}
+{-# LANGUAGE FlexibleInstances     #-}
+{-# LANGUAGE MultiParamTypeClasses #-}
+{-# LANGUAGE OverloadedStrings     #-}
+{-# LANGUAGE StandaloneDeriving    #-}
+{-# LANGUAGE TypeSynonymInstances  #-}
 module Skylighting.Tokenizer (
     tokenize
   , TokenizerConfig(..)
@@ -8,36 +14,24 @@ import Control.Applicative
 import Control.Monad.Except
 import Control.Monad.Reader
 import Control.Monad.State.Strict
-import qualified Data.ByteString.Char8 as BS
-import qualified Data.Map as Map
+import qualified Data.Attoparsec.ByteString.Char8 as A
 import Data.ByteString.Char8 (ByteString)
+import qualified Data.ByteString.Char8 as BS
+import qualified Data.ByteString.UTF8 as UTF8
 import Data.CaseInsensitive (mk)
-import Data.Char (isAlphaNum, isAscii, isLetter, isSpace, ord, isPrint)
+import Data.Char (isAlphaNum, isAscii, isLetter, isPrint, isSpace, ord)
+import qualified Data.Map as Map
 import Data.Maybe (catMaybes)
 import Data.Monoid
 import qualified Data.Set as Set
 import Data.Text (Text)
 import qualified Data.Text as Text
 import Data.Text.Encoding (decodeUtf8', encodeUtf8)
-import qualified Data.ByteString.UTF8 as UTF8
-import qualified Data.Attoparsec.ByteString.Char8 as A
 import Debug.Trace
 import Skylighting.Regex
 import Skylighting.Types
 import Text.Printf (printf)
 
-info :: String -> TokenizerM ()
-info s = do
-  tr <- asks traceOutput
-  when tr $ trace s (return ())
-
-infoContextStack :: TokenizerM ()
-infoContextStack = do
-  tr <- asks traceOutput
-  when tr $ do
-    ContextStack stack <- gets contextStack
-    info $ "CONTEXT STACK " ++ show (map cName stack)
-
 newtype ContextStack = ContextStack{ unContextStack :: [Context] }
   deriving (Show)
 
@@ -59,8 +53,106 @@ data TokenizerConfig = TokenizerConfig{
   , traceOutput :: Bool       -- ^ Generate trace output for debugging
 } deriving (Show)
 
-type TokenizerM =
-  ExceptT String (ReaderT TokenizerConfig (State TokenizerState))
+data Result e a = Success a
+                | Failure
+                | Error e
+     deriving (Functor)
+
+deriving instance (Show a, Show e) => Show (Result e a)
+
+data TokenizerM a = TM { runTokenizerM :: TokenizerConfig
+                                       -> TokenizerState
+                                       -> (TokenizerState, Result String a) }
+
+mapsnd :: (a -> b) -> (c, a) -> (c, b)
+mapsnd f (x, y) = (x, f y)
+
+instance Functor TokenizerM where
+  fmap f (TM g) = TM (\c s -> mapsnd (fmap f) (g c s))
+
+instance Applicative TokenizerM where
+  pure x = TM (\_ s -> (s, Success x))
+  (TM f) <*> (TM y) = TM (\c s ->
+                           case (f c s) of
+                              (s', Failure   ) -> (s', Failure)
+                              (s', Error e   ) -> (s', Error e)
+                              (s', Success f') ->
+                                  case (y c s') of
+                                    (s'', Failure   ) -> (s'', Failure)
+                                    (s'', Error e'  ) -> (s'', Error e')
+                                    (s'', Success y') -> (s'', Success (f' y')))
+
+
+instance Monad TokenizerM where
+  return = pure
+  (TM x) >>= f = TM (\c s ->
+                       case x c s of
+                            (s', Failure   ) -> (s', Failure)
+                            (s', Error e   ) -> (s', Error e)
+                            (s', Success x') -> g c s'
+                              where TM g = f x')
+
+instance Alternative TokenizerM where
+  empty = TM (\_ s -> (s, Failure))
+  (<|>) (TM x) (TM y) = TM (\c s ->
+                           case x c s of
+                                (_, Failure   )  -> y c s
+                                (s', Error e   ) -> (s', Error e)
+                                (s', Success x') -> (s', Success x'))
+  many (TM x) = TM (\c s ->
+                    case x c s of
+                       (_, Failure   )  -> (s, Success [])
+                       (s', Error e   ) -> (s', Error e)
+                       (s', Success x') -> mapsnd (fmap (x':)) (g c s')
+                         where TM g = many (TM x))
+  some x = (:) <$> x <*> many x
+
+instance MonadPlus TokenizerM where
+  mzero = empty
+  mplus = (<|>)
+
+instance MonadReader TokenizerConfig TokenizerM where
+  ask = TM (\c s -> (s, Success c))
+  local f (TM x) = TM (\c s -> x (f c) s)
+
+instance MonadState TokenizerState TokenizerM where
+  get = TM (\_ s -> (s, Success s))
+  put x = TM (\_ _ -> (x, Success ()))
+
+instance MonadError String TokenizerM where
+  throwError e = TM (\_ s -> (s, Error e))
+  catchError (TM x) f = TM (\c s -> case x c s of
+                                      (_, Error e) -> let TM y = f e in y c s
+                                      z            -> z)
+
+-- | Tokenize some text using 'Syntax'.
+tokenize :: TokenizerConfig -> Syntax -> Text -> Either String [SourceLine]
+tokenize config syntax inp =
+  case runTokenizerM action config initState of
+       (_, Success ls) -> Right ls
+       (_, Error e)    -> Left e
+       (_, Failure)    -> Left "Could not tokenize code"
+  where
+    action = mapM tokenizeLine (zip (BS.lines (encodeUtf8 inp)) [1..])
+    initState = startingState{ endline = Text.null inp
+                             , contextStack =
+                                   case lookupContext
+                                         (sStartingContext syntax) syntax of
+                                         Just c  -> ContextStack [c]
+                                         Nothing -> ContextStack [] }
+
+
+info :: String -> TokenizerM ()
+info s = do
+  tr <- asks traceOutput
+  when tr $ trace s (return ())
+
+infoContextStack :: TokenizerM ()
+infoContextStack = do
+  tr <- asks traceOutput
+  when tr $ do
+    ContextStack stack <- gets contextStack
+    info $ "CONTEXT STACK " ++ show (map cName stack)
 
 popContextStack :: TokenizerM ()
 popContextStack = do
@@ -110,20 +202,6 @@ lookupContext name syntax | Text.null name =
      else lookupContext (sStartingContext syntax) syntax
 lookupContext name syntax = Map.lookup name $ sContexts syntax
 
--- | Tokenize some text using 'Syntax'.
-tokenize :: TokenizerConfig -> Syntax -> Text -> Either String [SourceLine]
-tokenize config syntax inp =
-  evalState
-    (runReaderT
-      (runExceptT (mapM tokenizeLine $
-        zip (BS.lines $ encodeUtf8 inp) [1..])) config)
-    startingState{ input = encodeUtf8 inp
-                 , endline = Text.null inp
-                 , contextStack = case lookupContext
-                                       (sStartingContext syntax) syntax of
-                                       Just c  -> ContextStack [c]
-                                       Nothing -> ContextStack [] }
-
 startingState :: TokenizerState
 startingState =
   TokenizerState{ input = BS.empty
@@ -320,12 +398,16 @@ includeRules :: Maybe TokenType -> ContextName -> ByteString
 includeRules mbattr (syn, con) inp = do
   syntaxes <- asks syntaxMap
   case Map.lookup syn syntaxes >>= lookupContext con of
-       Nothing  -> throwError $ "Context lookup failed " ++ show (syn, con)
+       Nothing  -> do
+          cur <- currentContext
+          throwError $ "IncludeRules in " ++ Text.unpack (cSyntax cur) ++
+           " requires undefined context " ++
+           Text.unpack con ++ "##" ++ Text.unpack syn
        Just c   -> do
          mbtok <- msum (map (\r -> tryRule r inp) (cRules c))
          return $ case (mbtok, mbattr) of
                     (Just (NormalTok, xs), Just attr) -> Just (attr, xs)
-                    _ -> mbtok
+                    _                                 -> mbtok
 
 checkLineEnd :: Context -> TokenizerM ()
 checkLineEnd c = do
@@ -523,7 +605,7 @@ normalizeHighlighting ((t,x):xs)
 parseCStringChar :: ByteString -> TokenizerM Text
 parseCStringChar inp = do
   case A.parseOnly (A.match pCStringChar) inp of
-       Left _ -> mzero
+       Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
 pCStringChar :: A.Parser ()
@@ -539,7 +621,7 @@ pCStringChar = do
 parseCChar :: ByteString -> TokenizerM Text
 parseCChar inp = do
   case A.parseOnly (A.match pCChar) inp of
-       Left _ -> mzero
+       Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
 pCChar :: A.Parser ()
@@ -552,7 +634,7 @@ parseInt :: ByteString -> TokenizerM Text
 parseInt inp = do
   wordBoundary inp
   case A.parseOnly (A.match (pHex <|> pOct <|> pDec)) inp of
-       Left _ -> mzero
+       Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
 pDec :: A.Parser ()
@@ -565,7 +647,7 @@ parseOct :: ByteString -> TokenizerM Text
 parseOct inp = do
   wordBoundary inp
   case A.parseOnly (A.match pHex) inp of
-       Left _ -> mzero
+       Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
 pOct :: A.Parser ()
@@ -580,7 +662,7 @@ parseHex :: ByteString -> TokenizerM Text
 parseHex inp = do
   wordBoundary inp
   case A.parseOnly (A.match pHex) inp of
-       Left _ -> mzero
+       Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r) -- assumes ascii
 
 pHex :: A.Parser ()
@@ -595,7 +677,7 @@ guardWordBoundary :: A.Parser ()
 guardWordBoundary = do
   mbw <- A.peekChar
   case mbw of
-       Just c ->  guard $ isWordBoundary '0' c
+       Just c  ->  guard $ isWordBoundary '0' c
        Nothing -> return ()
 
 mbMinus :: A.Parser ()
@@ -608,7 +690,7 @@ parseFloat :: ByteString -> TokenizerM Text
 parseFloat inp = do
   wordBoundary inp
   case A.parseOnly (A.match pFloat) inp of
-       Left _ -> mzero
+       Left _      -> mzero
        Right (r,_) -> takeChars (BS.length r)  -- assumes all ascii
   where pFloat :: A.Parser ()
         pFloat = do
@@ -621,8 +703,8 @@ parseFloat inp = do
                                          mbPlusMinus >> digits)
           mbnext <- A.peekChar
           case mbnext of
-               Nothing   -> return ()
-               Just c    -> guard (not $ A.inClass "." c)
+               Nothing -> return ()
+               Just c  -> guard (not $ A.inClass "." c)
           guard $ (before && not dot && e)     -- 5e2
                || (before && dot && (after || not e)) -- 5.2e2 or 5.2 or 5.
                || (not before && dot && after) -- .23 or .23e2
